"""
based on
https://github.com/gwastro/pycbc/blob/master/pycbc/filter/matchedfilter.py
"""

import tensorflow as tf
import numpy as np
from scrinet.analysis.frequencyseries import FrequencySeries, fftfreq
from scrinet.analysis.timeseries import TimeSeries
from scrinet.analysis.tf_fft import ifft
from math import frexp


def ceilpow2(n):
    """convenience function to determine a power-of-2 upper frequency limit"""
    signif, exponent = frexp(n)
    if (signif < 0):
        return 1
    if (signif == 0.5):
        exponent -= 1
    return (1) << exponent


def inner(x, y):
    """ Return the inner product of the array with complex conjugation.
    """
    return tf.reduce_sum(tf.math.conj(x) * y)


def weighted_inner(x, y, weight):
    """ Return the inner product of the array with complex conjugation.
    """
    return tf.reduce_sum(tf.math.conj(x) * y / weight)


def correlate(x, y):
    """
    computes the complex conjuate of x multiplied by y
    Parameters
    ----------
    x : tf.Tensor
    y : tf.Tensor
    Returns
    -------
    tf.Tensor:
        the result of conj(x)*y
    """
    return tf.math.conj(x) * y


def make_frequency_series(vec):
    """Return a frequency series of the input vector.
    If the input is a frequency series it is returned, else if the input
    vector is a real time series it is fourier transformed and returned as a
    frequency series.
    Parameters
    ----------
    vector : TimeSeries or FrequencySeries
    Returns
    -------
    Frequency Series: FrequencySeries
        A frequency domain version of the input vector.
    """
    if isinstance(vec, FrequencySeries):
        return vec
    if isinstance(vec, TimeSeries):
        vectilde = vec.to_frequencyseries()
        return vectilde
    else:
        raise TypeError("Can only convert a TimeSeries to a FrequencySeries")


def get_cutoff_indices(flow, fhigh, df, N):
    """
    Gets the indices of a frequency series at which to stop an overlap
    calculation.
    Parameters
    ----------
    flow: float
        The frequency (in Hz) of the lower index.
    fhigh: float
        The frequency (in Hz) of the upper index.
    df: float
        The frequency step (in Hz) of the frequency series.
    N: int
        The number of points in the **time** series. Can be odd
        or even.
    Returns
    -------
    kmin: int
    kmax: int
    """
    if flow:
        kmin = int(flow / df)
        if kmin < 0:
            err_msg = "Start frequency cannot be negative. "
            err_msg += "Supplied value and kmin {} and {}".format(flow, kmin)
            raise ValueError(err_msg)
    else:
        kmin = 1
    if fhigh:
        kmax = int(fhigh / df)
        if kmax > int((N + 1) / 2.):
            kmax = int((N + 1) / 2.)
    else:
        # int() truncates towards 0, so this is
        # equivalent to the floor of the float
        kmax = int((N + 1) / 2.)

    if kmax <= kmin:
        err_msg = "Kmax cannot be less than or equal to kmin. "
        err_msg += "Provided values of freqencies (min,max) were "
        err_msg += "{} and {} ".format(flow, fhigh)
        err_msg += "corresponding to (kmin, kmax) of "
        err_msg += "{} and {}.".format(kmin, kmax)
        raise ValueError(err_msg)

    return kmin, kmax


def sigmasq(htilde, psd=None, low_frequency_cutoff=None,
            high_frequency_cutoff=None):
    """Return the loudness of the waveform. This is defined (see Duncan
    Brown's thesis) as the unnormalized matched-filter of the input waveform,
    htilde, with itself. This quantity is usually referred to as (sigma)^2
    and is then used to normalize matched-filters with the data.
    Parameters
    ----------
    htilde : TimeSeries or FrequencySeries
        The input vector containing a waveform.
    psd : {None, FrequencySeries}, optional
        The psd used to weight the accumulated power.
    low_frequency_cutoff : {None, float}, optional
        The frequency to begin considering waveform power.
    high_frequency_cutoff : {None, float}, optional
        The frequency to stop considering waveform power.
    Returns
    -------
    sigmasq: tf.float32 or tf.float64
    """
    htilde = make_frequency_series(htilde)
    N = (htilde.data.shape[0] - 1) * 2
    norm = 4.0 * htilde.delta_f
    kmin, kmax = get_cutoff_indices(low_frequency_cutoff,
                                    high_frequency_cutoff, htilde.delta_f, N)

    htilde_cut = FrequencySeries(htilde.data[kmin:kmax],
                                 delta_f=htilde.delta_f, epoch=htilde._epoch)

    if psd is not None:
        try:
            np.testing.assert_almost_equal(htilde_cut.delta_f, psd.delta_f)
        except AssertionError:
            raise ValueError('Waveform does not have same delta_f as psd')

    if psd is None:
        sq = inner(htilde_cut.data, htilde_cut.data)
    else:
        sq = weighted_inner(
            htilde_cut.data, htilde_cut.data, psd.data[kmin:kmax])

    return tf.math.real(sq) * norm


# @tf.function
def matched_filter_core(template, data, psd=None, low_frequency_cutoff=None,
                        high_frequency_cutoff=None):
    """ Return the complex snr and normalization.
    Return the complex snr, along with its associated normalization of the template,
    matched filtered against the data.
    Parameters
    ----------
    template : TimeSeries or FrequencySeries
        The template waveform
    data : TimeSeries or FrequencySeries
        The strain data to be filtered.
    psd : {FrequencySeries}, optional
        The noise weighting of the filter.
    low_frequency_cutoff : {None, float}, optional
        The frequency to begin the filter calculation. If None, begin at the
        first frequency after DC.
    high_frequency_cutoff : {None, float}, optional
        The frequency to stop the filter calculation. If None, continue to the
        the nyquist frequency.
    Returns
    -------
    snr : TimeSeries
        A time series containing the complex snr.
    corrrelation: FrequencySeries
        A frequency series containing the correlation vector.
    norm : float
        The normalization of the complex snr.
    """
    htilde = make_frequency_series(template)
    stilde = make_frequency_series(data)

    if htilde.data.shape[0] != stilde.data.shape[0]:
        raise ValueError("Length of template and data must match")

    N = (stilde.data.shape[0] - 1) * 2
    kmin, kmax = get_cutoff_indices(low_frequency_cutoff,
                                    high_frequency_cutoff, stilde.delta_f, N)

    below_kmin = tf.zeros(kmin, dtype=htilde._dtype)
    after_kmax = tf.zeros(N - kmax, dtype=htilde._dtype)
    within_k = correlate(
        htilde.data[kmin:kmax], stilde.data[kmin:kmax])

    qtilde = tf.concat([below_kmin, within_k, after_kmax], axis=0)
    # correlation array
    # qtilde_list = tf.unstack(tf.zeros(N, dtype=tf.dtypes.as_dtype(stilde.dtype)))
    #
    # # qtilde_np = qtilde.numpy()
    #
    # print(htilde.data[kmin:kmax])
    # print(stilde.data[kmin:kmax])
    #
    # print(correlate(
    #     htilde.data[kmin:kmax], stilde.data[kmin:kmax]))
    #
    # print('correlate worked')
    # qtilde_list[kmin:kmax] = correlate(
    #     htilde.data[kmin:kmax], stilde.data[kmin:kmax])
    #
    # print('list been made')
    # qtilde = tf.stack(qtilde_list)

    # print('list stacked into tensor')

    # qtilde = tf.convert_to_tensor(qtilde_np, dtype=qtilde.dtype)

    # qtilde[kmin:kmax] = correlate(htilde[kmin:kmax], stilde[kmin:kmax])

    if psd is not None:
        if isinstance(psd, FrequencySeries):
            try:
                np.testing.assert_almost_equal(stilde.delta_f, psd.delta_f)
            except AssertionError:
                raise ValueError("PSD delta_f does not match data")
            tmp_a = qtilde.numpy()
            tmp_b = psd.data.numpy()
            tmp_a[kmin:kmax] /= tmp_b[kmin:kmax]
            qtilde = tf.convert_to_tensor(tmp_a, dtype=qtilde._dtype)
        else:
            raise TypeError("PSD must be a FrequencySeries")

    qtilde_fs = FrequencySeries(
        qtilde, delta_f=stilde.delta_f, epoch=stilde._epoch)

    # we don't want to normlise the ifft here.. not sure exactly why
    q = ifft(qtilde_fs, normalise=False)

    h_norm = sigmasq(htilde, psd, low_frequency_cutoff, high_frequency_cutoff)

    norm = (4.0 * stilde.delta_f) / tf.math.sqrt(h_norm)

    return q, qtilde_fs, norm


def matched_filter(template, data, psd=None, low_frequency_cutoff=None,
                   high_frequency_cutoff=None):
    """ Return the complex snr.
    Return the complex snr, along with its associated normalization of the
    template, matched filtered against the data.
    Parameters
    ----------
    template : TimeSeries or FrequencySeries
        The template waveform
    data : TimeSeries or FrequencySeries
        The strain data to be filtered.
    psd : FrequencySeries
        The noise weighting of the filter.
    low_frequency_cutoff : {None, float}, optional
        The frequency to begin the filter calculation. If None, begin at the
        first frequency after DC.
    high_frequency_cutoff : {None, float}, optional
        The frequency to stop the filter calculation. If None, continue to the
        the nyquist frequency.
    Returns
    -------
    snr : TimeSeries
        A time series containing the complex snr.
    """
    snr, _, norm = matched_filter_core(template, data, psd=psd,
                                       low_frequency_cutoff=low_frequency_cutoff,
                                       high_frequency_cutoff=high_frequency_cutoff)

    snr.data = snr.data * tf.cast(norm, tf.dtypes.as_dtype(snr._dtype))

    return snr


def match(vec1, vec2, psd=None, low_frequency_cutoff=None,
          high_frequency_cutoff=None):
    """ Return the match between the two TimeSeries or FrequencySeries.
    Return the match between two waveforms. This is equivalent to the overlap
    maximized over time and phase.
    Parameters
    ----------
    vec1 : TimeSeries or FrequencySeries
        The input vector containing a waveform.
    vec2 : TimeSeries or FrequencySeries
        The input vector containing a waveform.
    psd : Frequency Series
        A power spectral density to weight the overlap.
    low_frequency_cutoff : {None, float}, optional
        The frequency to begin the match.
    high_frequency_cutoff : {None, float}, optional
        The frequency to stop the match.
    Returns
    -------
    match: float
    """

    htilde = make_frequency_series(vec1)
    stilde = make_frequency_series(vec2)

    snr, _, snr_norm = matched_filter_core(htilde, stilde, psd, low_frequency_cutoff,
                                           high_frequency_cutoff)

    max_idx = tf.argmax(tf.math.abs(snr.data))
    maxsnr = tf.math.abs(snr.data[max_idx])
    v2_norm = sigmasq(stilde, psd, low_frequency_cutoff, high_frequency_cutoff)
    return maxsnr * snr_norm / tf.math.sqrt(v2_norm)


def coalign_waveforms(h1, h2, psd=None,
                      low_frequency_cutoff=None,
                      high_frequency_cutoff=None,
                      resize=True):
    """ Return two time series which are aligned in time and phase.
    The alignment is only to the nearest sample point and all changes to the
    phase are made to the first input waveform. Waveforms should not be split
    accross the vector boundary. If it is, please use roll or cyclic time shift
    to ensure that the entire signal is contiguous in the time series.
    Parameters
    ----------
    h1: scrinet.analysis.timeseries.TimeSeries
        The first waveform to align.
    h2: scrinet.analysis.timeseries.TimeSeries
        The second waveform to align.
    psd: {None, scrinet.analysis.frequencyseries.FrequencySeries}
        A psd to weight the alignment
    low_frequency_cutoff: {None, float}
        The low frequency cutoff to weight the matching in Hz.
    high_frequency_cutoff: {None, float}
        The high frequency cutoff to weight the matching in Hz.
    resize: Optional, {True, boolean}
        If true, the vectors will be resized to match each other. If false,
        they must be the same length and even in length
    Returns
    -------
    h1: scrinet.analysis.timeseries.TimeSeries
        The shifted waveform to align with h2
    h2: scrinet.analysis.timeseries.TimeSeries
        The resized (if necessary) waveform to align with h1.
    """
    if resize:
        # need to zero pad to same length
        h1_shape = h1.data.shape[0]
        h2_shape = h2.data.shape[0]
        new_len = np.max([h1_shape, h2_shape])
        new_len = ceilpow2(new_len)

        # arr_to_append = np.zeros(np.abs(new_len-h1_shape))
        # h1 = TimeSeries(np.append(h1.data.numpy(), arr_to_append), delta_t=h1.delta_t, epoch=h1._epoch)
        arr_to_append = tf.zeros(tf.math.abs(
            new_len - h1_shape), dtype=h1._dtype)
        h1 = TimeSeries(tf.concat(
            [h1.data, arr_to_append], axis=0), delta_t=h1.delta_t, epoch=h1._epoch)

        # arr_to_append = np.zeros(np.abs(new_len-h2_shape))
        # h2 = TimeSeries(np.append(h2.data.numpy(), arr_to_append), delta_t=h2.delta_t, epoch=h2._epoch)

        arr_to_append = tf.zeros(tf.math.abs(
            new_len - h2_shape), dtype=h2._dtype)
        h2 = TimeSeries(tf.concat(
            [h2.data, arr_to_append], axis=0), delta_t=h2.delta_t, epoch=h2._epoch)

    elif h1.data.shape[0] != h2.data.shape[0]:
        raise ValueError("Time series must be the same size and even if you do "
                         "not allow resizing")

    snr = matched_filter(h1, h2, psd=psd,
                         low_frequency_cutoff=low_frequency_cutoff,
                         high_frequency_cutoff=high_frequency_cutoff)

    max_idx = tf.argmax(tf.math.abs(snr.data))

    norm = tf.cast(tf.math.abs(snr.data[max_idx]), snr._dtype)

    rotation = snr.data[max_idx] / norm

    h1_tilde = h1.to_frequencyseries()
    h1_tilde_data = h1.to_frequencyseries().data * rotation
    h1_tilde = FrequencySeries(
        h1_tilde_data, delta_f=h1_tilde.delta_f, epoch=h1_tilde._epoch)

    h1 = h1_tilde.to_timeseries()

    rolled_data = tf.roll(h1.data, max_idx, axis=0)

    h1 = TimeSeries(tf.math.real(rolled_data),
                    delta_t=h2.delta_t, epoch=h2.start_time)
    return h1, h2



from pycondor import Job, Dagman
from distutils.spawn import find_executable
from os.path import abspath, join
import configparser


class Node(Job):
    def add_args_from_cp(self, cp, section, set_name=None):
        """
        Parse command line options from a given section in an ini file and
        pass to the executable.
        cp: ConfigParser object
            ConfigParser object pointing to the ini file.
        section: str
            section of the ini file to add to the options.
        set_name: str
            subection name. If not None then will look for
            options of the form f"{section}-{set_name}"
        """
        args = []
        for opt in cp.options(section):
            arg = (cp.get(section, opt)).strip()
            args.append(f"--{opt} {arg}")

        if set_name:
            name = f"{section}-{set_name}"
            for opt in cp.options(name):
                arg = (cp.get(name, opt)).strip()
                args.append(f"--{opt} {arg}")
        self.add_arg(arg=" ".join(args))

    def get_request_memory(self, cp, name):
        """

        Parameters
        ----------
        cp: configparser.ConfigParser
        name: str
            name of subsection
        'request_memory' should always be in [workflow]
        but it can be overridden in an executable specific
        subsection such as [workflow-genwf-seed]
        """
        try:
            request_memory = cp.get("workflow-"+name, "request_memory")
        except:
            request_memory = cp.get("workflow", "request_memory")
        return request_memory


class GenWaveformJob(Node):
    """Waveform generator Job

    Parameters
    ----------
    cp: configparser.ConfigParser
    section: str
        name of section in cp.
        Also name of executable key in [executables]
    set_name: str
        subsection name.
        either 'seed', 'train', 'validation' or 'test'
    idx: int
        index to append to output directory
        Default is None and won't append anything
    """

    def __init__(self, cp, section, set_name, idx=None):
        name = f"{section}-{set_name}"

        executable = cp.get("executables", section)
        executable = find_executable(executable)

        log_dir = cp.get("workflow", "log_dir")

        request_memory = self.get_request_memory(cp, name)

        request_cpus = cp.get(name, "n-cores")

        accounting_group = cp.get("workflow", "accounting_group")
        extra_lines = [f'accounting_group = {accounting_group}']

        if idx:
            submit_name = name + f"-{idx}"
        else:
            submit_name = name

        super().__init__(
            name=submit_name,
            executable=executable,
            error=abspath(join(log_dir, 'error')),
            log=abspath(join(log_dir, 'log')),
            output=abspath(join(log_dir, 'output')),
            submit=abspath(join(log_dir, 'submit')),
            request_memory=request_memory,
            request_cpus=request_cpus,
            getenv=True,
            universe="vanilla",
            extra_lines=extra_lines
        )

        self.add_args_from_cp(cp, section, set_name, idx)

    def add_args_from_cp(self, cp, section, set_name=None, idx=None):
        """
        overrides method to include the idx variable

        Parse command line options from a given section in an ini file and
        pass to the executable.
        cp: ConfigParser object
            ConfigParser object pointing to the ini file.
        section: str
            section of the ini file to add to the options.
        set_name: str
            subection name. If not None then will look for
            options of the form f"{section}-{set_name}"
        idx: int
            index to append to output directory ("--output-dir" option)
            Default is None and won't append anything
        """
        args = []
        for opt in cp.options(section):
            arg = (cp.get(section, opt)).strip()
            args.append(f"--{opt} {arg}")

        if set_name:
            name = f"{section}-{set_name}"
            for opt in cp.options(name):
                arg = (cp.get(name, opt)).strip()
                if opt == "output-dir":
                    if idx:
                        arg = arg+f"_{idx}"
                args.append(f"--{opt} {arg}")
        self.add_arg(arg=" ".join(args))


class CombineWaveformJob(Node):
    """Combine Waveform Job

    Parameters
    ----------
    cp: configparser.ConfigParser
    section: str
        name of section in cp.
        Also name of executable key in [executables]
    set_name: str
        subsection name.
        either 'seed', 'train', 'validation' or 'test'
    """

    def __init__(self, cp, section, set_name):
        name = f"{section}-{set_name}"

        executable = cp.get("executables", section)
        executable = find_executable(executable)

        log_dir = cp.get("workflow", "log_dir")

        request_memory = self.get_request_memory(cp, name)

        request_cpus = 1

        accounting_group = cp.get("workflow", "accounting_group")
        extra_lines = [f'accounting_group = {accounting_group}']

        super().__init__(
            name=name,
            executable=executable,
            error=abspath(join(log_dir, 'error')),
            log=abspath(join(log_dir, 'log')),
            output=abspath(join(log_dir, 'output')),
            submit=abspath(join(log_dir, 'submit')),
            request_memory=request_memory,
            request_cpus=request_cpus,
            getenv=True,
            universe="vanilla",
            extra_lines=extra_lines
        )

        self.add_args_from_cp(cp, section, set_name)


class BuildReducedBasisJob(Node):
    """Build reduced basis Job

    Parameters
    ----------
    cp: configparser.ConfigParser
    section: str
        name of section in cp.
        Also name of executable key in [executables]
    set_name: str
        subsection name.
        either 'seed', 'train', 'validation' or 'test'
    """

    def __init__(self, cp, section, set_name):
        name = f"{section}-{set_name}"

        executable = cp.get("executables", section)
        executable = find_executable(executable)

        log_dir = cp.get("workflow", "log_dir")

        request_memory = self.get_request_memory(cp, name)

        request_cpus = 1

        accounting_group = cp.get("workflow", "accounting_group")
        extra_lines = [f'accounting_group = {accounting_group}']

        super().__init__(
            name=name,
            executable=executable,
            error=abspath(join(log_dir, 'error')),
            log=abspath(join(log_dir, 'log')),
            output=abspath(join(log_dir, 'output')),
            submit=abspath(join(log_dir, 'submit')),
            request_memory=request_memory,
            request_cpus=request_cpus,
            getenv=True,
            universe="vanilla",
            extra_lines=extra_lines
        )

        self.add_args_from_cp(cp, section, set_name)


class GenTrainingSetJob(Node):
    """Generate training set Job

    Parameters
    ----------
    cp: configparser.ConfigParser
    section: str
        name of section in cp.
        Also name of executable key in [executables]
    data: str
        name of data.
        'amp', 'phase', 'real', 'imag', 'freq'
    set_name: str
        subsection name.
        either 'train' or 'val'
    """

    def __init__(self, cp, section, data, set_name):
        data_set_name = f"{data}-{set_name}"
        name = f"{section}-{data_set_name}"

        executable = cp.get("executables", section)
        executable = find_executable(executable)

        log_dir = cp.get("workflow", "log_dir")

        request_memory = self.get_request_memory(cp, name)

        request_cpus = 1

        accounting_group = cp.get("workflow", "accounting_group")
        extra_lines = [f'accounting_group = {accounting_group}']

        super().__init__(
            name=name,
            executable=executable,
            error=abspath(join(log_dir, 'error')),
            log=abspath(join(log_dir, 'log')),
            output=abspath(join(log_dir, 'output')),
            submit=abspath(join(log_dir, 'submit')),
            request_memory=request_memory,
            request_cpus=request_cpus,
            getenv=True,
            universe="vanilla",
            extra_lines=extra_lines
        )

        self.add_args_from_cp(cp, section, data_set_name)


class FitJob(Node):
    """Fit Job

    Parameters
    ----------
    cp: configparser.ConfigParser
    section: str
        name of section in cp.
        Also name of executable key in [executables]
    data: str
        subsection name.
        either 'amp', 'phase', 'real', 'imag', 'freq'
    """

    def __init__(self, cp, section, data):
        name = f"{section}-{data}"

        executable = cp.get("executables", section)
        executable = find_executable(executable)

        log_dir = cp.get("workflow", "log_dir")

        request_memory = self.get_request_memory(cp, name)

        accounting_group = cp.get("workflow", "accounting_group")
        extra_lines = [f'accounting_group = {accounting_group}']

        request_cpus = cp.get("workflow-"+name, "request_cpus")
        if request_cpus == str(0):
            raise ValueError(
                "You have requested 0 cpus. Please request at least 1")

        try:
            request_gpus = cp.get("workflow-"+name, "request_gpus")
            extra_lines.append(f"request_gpus = {request_gpus}")
        except:
            pass
        super().__init__(
            name=name,
            executable=executable,
            error=abspath(join(log_dir, 'error')),
            log=abspath(join(log_dir, 'log')),
            output=abspath(join(log_dir, 'output')),
            submit=abspath(join(log_dir, 'submit')),
            request_memory=request_memory,
            request_cpus=request_cpus,
            getenv=True,
            universe="vanilla",
            extra_lines=extra_lines
        )

        self.add_args_from_cp(cp, section, data)


class EvaluateJob(Node):
    """Evaluate Job

    Parameters
    ----------
    cp: configparser.ConfigParser
    section: str
        name of section in cp.
        Also name of executable key in [executables]
    set_name: str
        subsection name.
        either 'train', 'validation' or 'test'
    """

    def __init__(self, cp, section, set_name):
        name = f"{section}-{set_name}"

        executable = cp.get("executables", section)
        executable = find_executable(executable)

        log_dir = cp.get("workflow", "log_dir")

        request_memory = self.get_request_memory(cp, name)

        accounting_group = cp.get("workflow", "accounting_group")
        extra_lines = [f'accounting_group = {accounting_group}']

        request_cpus = 1

        super().__init__(
            name=name,
            executable=executable,
            error=abspath(join(log_dir, 'error')),
            log=abspath(join(log_dir, 'log')),
            output=abspath(join(log_dir, 'output')),
            submit=abspath(join(log_dir, 'submit')),
            request_memory=request_memory,
            request_cpus=request_cpus,
            getenv=True,
            universe="vanilla",
            extra_lines=extra_lines
        )

        self.add_args_from_cp(cp, section, set_name)


class MakeWebpageJob(Node):
    """Make Webpage Job

    Parameters
    ----------
    cp: configparser.ConfigParser
    section: str
        name of section in cp.
        Also name of executable key in [executables]
    """

    def __init__(self, cp, section):
        name = f"{section}"

        executable = cp.get("executables", section)
        executable = find_executable(executable)

        log_dir = cp.get("workflow", "log_dir")

        request_memory = self.get_request_memory(cp, name)

        accounting_group = cp.get("workflow", "accounting_group")
        extra_lines = [f'accounting_group = {accounting_group}']

        request_cpus = 1

        super().__init__(
            name=name,
            executable=executable,
            error=abspath(join(log_dir, 'error')),
            log=abspath(join(log_dir, 'log')),
            output=abspath(join(log_dir, 'output')),
            submit=abspath(join(log_dir, 'submit')),
            request_memory=request_memory,
            request_cpus=request_cpus,
            getenv=True,
            universe="vanilla",
            extra_lines=extra_lines
        )

        self.add_args_from_cp(cp, section)



import tensorflow as tf
from scrinet.analysis import frequencyseries
from scrinet.analysis import timeseries


def fft(y):
    """ Fourier transform of y.
    This is fixed to assume:
        1. y is a real TimeSeries
        2. we use a real fft i.e. rfft
    Parameters
    ----------
    y : TimeSeries
        The input vector.
    Returns
    -------
    FrequencySeries:
        The (real) fourier transform of this (real) time series.
    """
    assert isinstance(y, timeseries.TimeSeries)
    f = tf.signal.rfft(y.data)
    dt = tf.cast(y.delta_t, f.dtype)
    f = f * dt
    delta_f = 1.0/(y.delta_t * y.data.shape[0])
    fs = frequencyseries.FrequencySeries(
        f, delta_f=delta_f, epoch=y.start_time)
    return fs


def ifft(ytilde, normalise=True):
    """ Inverse fourier transform of ytilde.
    This is fixed to assume:
        1. ytilde is a complex FrequencySeries
        2. we use a complex ifft i.e. ifft
    Parameters
    ----------
    ytilde : FrequencySeries
        The input vector.
    normalise : {bool, True}
        The output is always multiplied by the length of the ytilde data.
        If true then the output is also multiplied by delta_f * 2
    Returns
    -------
    TimeSeries:
        The inverse fourier transform of this frequency series.
    """
    assert isinstance(ytilde, frequencyseries.FrequencySeries)
    t = tf.signal.ifft(ytilde.data)

    norm = ytilde.data.shape[0]
    df = tf.cast(ytilde.delta_f, t.dtype)
    if normalise:
        norm *= df * 2
    t = t * norm
    delta_t = 1.0/(ytilde.delta_f * ytilde.data.shape[0])
    return timeseries.TimeSeries(t, delta_t=delta_t, epoch=ytilde._epoch)



import logging
import sys
import os
import numpy as np
import h5py

from scrinet.fits.nn import RegressionANN
from scrinet.interfaces import rotations


def init_logger():
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

    # need to do this to avoid print things twice
    # https://stackoverflow.com/a/6729713/12840171
    if not logger.handlers:
        handler = logging.StreamHandler(sys.stdout)
        handler.flush = sys.stdout.flush
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    else:
        logger.handlers[0].formatter = formatter
        logger.setLevel(logging.INFO)
        logger.handlers[0].flush()

    return logger


def load_data(data_to_model, dir_name, start_index=0, end_index=None):
    with h5py.File(os.path.join(dir_name, 'times.h5'), 'r') as f:
        x = f['times'][:]

    with h5py.File(os.path.join(dir_name, 'coords.h5'), 'r') as f:
        if end_index is None:
            end_index = f['data'].shape[0]
        coords = f['data'][start_index:end_index]

    with h5py.File(os.path.join(dir_name, f'{data_to_model}.h5'), 'r') as f:
        if end_index is None:
            end_index = f['data'].shape[0]
        data = f['data'][start_index:end_index]

    return x, data, coords


def load_greedy_points(filename):
    with h5py.File(filename, 'r') as f:
        coords = f['data'][:]

    return coords


def load_model(basis_file, nn_weights_file, X_scalers_file, Y_scalers_file):
    model = RegressionANN()
    model.load_model(nn_weights_file)
    model.load_X_scalers(X_scalers_file)
    model.load_Y_scalers(Y_scalers_file)
    model.scaleX = True
    model.scaleY = True

    basis = np.load(basis_file)

    return model, basis


def wave_sur_many(pars, amp_model, amp_basis, phase_model, phase_basis):
    """
    pars: np.ndarray of pars to evaluate model
        pars.shape = (N,M) where N is the number of waveforms
            and M is the dimensionality.
            M=0 should always be mass-ratio
    """
    # pars = np.array(list(zip(qs, chis)))
    # pars = np.array(list(zip(qs, chi1xs, chi1zs)))
    pars = pars.copy()
    pars[:, 0] = np.log(pars[:, 0])  # need to log the mass-ratio

    amp_alpha = amp_model.predict(pars)
    amp = np.dot(amp_alpha, amp_basis)

    phase_alpha = phase_model.predict(pars)
    phase = np.dot(phase_alpha, phase_basis)

    h = amp * np.exp(-1.j * phase)

    return np.real(h), np.imag(h), amp, phase


def real_imag_wave_sur_many(pars, real_model, real_basis, imag_model, imag_basis):
    """
    Surrogate generator function if modelling real and imag parts of hlm
    pars: np.ndarray of pars to evaluate model
        pars.shape = (N,M) where N is the number of waveforms
            and M is the dimensionality.
            M=0 should always be mass-ratio
    returns:
        hreal, himag: np.ndarray
    """
    # pars = np.array(list(zip(qs, chis)))
    # pars = np.array(list(zip(qs, chi1xs, chi1zs)))
    pars = pars.copy()
    pars[:, 0] = np.log(pars[:, 0])  # need to log the mass-ratio

    real_alpha = real_model.predict(pars)
    hreal = np.dot(real_alpha, real_basis)

    imag_alpha = imag_model.predict(pars)
    himag = np.dot(imag_alpha, imag_basis)

    return hreal, himag


def coprec_wave_sur_many(pars, x, alpha0, thetaJN, phi0, amp_model, amp_basis, phase_model, phase_basis, alpha_model, alpha_basis, beta_model, beta_basis, gamma_model, gamma_basis):
    """
    pars: np.ndarray of pars to evaluate model
        pars.shape = (N,M) where N is the number of waveforms
            and M is the dimensionality.
            M=0 should always be mass-ratio
    """
    # pars = np.array(list(zip(qs, chis)))
    # pars = np.array(list(zip(qs, chi1xs, chi1zs)))
    pars = pars.copy()
    pars[:, 0] = np.log(pars[:, 0])  # need to log the mass-ratio

    amp_alpha = amp_model.predict(pars)
    amp = np.dot(amp_alpha, amp_basis)

    phase_alpha = phase_model.predict(pars)
    phase = np.dot(phase_alpha, phase_basis)

    N, M = amp.shape

    hlms = []
    for i in range(N):
        modes_dict = {
            (2, 2): amp[i]*np.exp(1.j*phase[i]),
            (2, 1): np.zeros(M),
            (2, 0): np.zeros(M),
            (2, -1): np.zeros(M),
            (2, -2): amp[i]*np.exp(-1.j*phase[i])
        }
        hlms.append(modes_dict)

    alpha_alpha = alpha_model.predict(pars)
    alpha = np.dot(alpha_alpha, alpha_basis)

    beta_alpha = beta_model.predict(pars)
    beta = np.dot(beta_alpha, beta_basis)

    gamma_alpha = gamma_model.predict(pars)
    gamma = np.dot(gamma_alpha, gamma_basis)

    h = []
    for i in range(N):
        wr = rotations.WaveformRotations(
            x, hlms[i], frame='coprec', alpha=alpha[i], beta=beta[i], gamma=gamma[i])
        wr.set_alpha0_thetaJN_phi0(alpha0[i], thetaJN[i], phi0[i])
        wr.from_coprecessing_frame_to_inertial_frame()
        h.append(wr.hlms[(2, 2)])

    return np.real(np.array(h)), np.imag(np.array(h))


def match(h1, h2, times):

    dt = times[1] - times[0]
    n = len(times)
    df = 1.0/(n*dt)
    norm = 4. * df

    h1_fft = np.fft.fft(h1)
    h2_fft = np.fft.fft(h2)

    h1h1_sq = np.vdot(h1_fft, h1_fft) * norm
    h2h2_sq = np.vdot(h2_fft, h2_fft) * norm

    h1h1 = dt * np.sqrt(h1h1_sq)
    h2h2 = dt * np.sqrt(h2h2_sq)

    ifft = np.fft.ifft(np.conj(h1_fft) * h2_fft)

    return ifft / h1h1 / h2h2 * 4 * dt



import tensorflow as tf
from scrinet.fits import nn
import numpy as np
import arviz as az


def load_amp_model(path_to_amp_model: str = 'amp/'):
    """
    helper to load Fitted scrinet Neural Network for amplitude basis
    :param path_to_amp_model: Location of amp model to be used
    :return:
    """
    model = nn.RegressionANN()
    model.load_model(path_to_amp_model + "best.h5")
    model.load_X_scalers(path_to_amp_model + "X_scalers.npy")
    model.load_Y_scalers(path_to_amp_model + "Y_scalers.npy")
    model.scaleX = True
    model.scaleY = True

    basis = np.load(path_to_amp_model + "amp_basis.npy")

    return model, basis


def load_phase_model(path_to_phase_model: str = 'phase/'):
    """
    helper to load Fitted scrinet Neural Network for phase basis

    :param path_to_phase_model: Location of phase model to be used
    :return:
    """
    model = nn.RegressionANN()
    model.load_model(path_to_phase_model + "best.h5")
    model.load_X_scalers(path_to_phase_model + "X_scalers.npy")
    model.load_Y_scalers(path_to_phase_model + "Y_scalers.npy")
    model.scaleX = True
    model.scaleY = True

    basis = np.load(path_to_phase_model + "phase_basis.npy")

    return model, basis


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
@tf.function
def predict_hack(model, _input):
    """
    Function to re-create model.predict(input) but allows the graph to be built correctly
    (see https://github.com/tensorflow/tensorflow/issues/33997)

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to make a prediction
    :return:
    """
    x = _input
    x = x_scale_hack(model, x)
    pred = tf.convert_to_tensor(x)
    pred = tf.cast(pred, tf.float64)
    # pred = x * tf.ones([len(x)], dtype=tf.float64)
    # this should be an argument for ndim of problem
    pred = tf.reshape(pred, shape=(-1, 3))
    for layer in model.model.layers:
        pred = layer(pred)
    pred = tf.reshape(
        pred, shape=(-1, model.model.get_output_shape_at(-1)[1],))
    pred = y_inv_scale_hack(model, pred)
    return pred


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
# @tf.function
def x_scale_hack(model, _input):
    """
    Function to re-create sklearn standard scaler for the input (as a tensorflow function)

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    x = _input
    means = tf.constant([model.X_scalers[i].mean_[0]
                         for i in range(x.shape[0])])
    stds = tf.constant([model.X_scalers[i].scale_[0]
                        for i in range(x.shape[0])])
    means = tf.cast(means, tf.float64)
    stds = tf.cast(stds, tf.float64)

    x_scaled = (tf.reshape(x, shape=(-1, x.shape[0],)) - means) / stds
    x_scaled = tf.cast(x_scaled, tf.float32)
    return x_scaled


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
# @tf.function
def get_model_y_std_scalers(model, _input):
    """
    Re-casts the sklearn output std scalers for each output basis

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    std = [model.Y_scalers[i].scale_[0] for i in range(_input.shape[1])]
    std = tf.stack(std, axis=0)
    std = tf.cast(std, tf.float32)
    return std


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
# @tf.function
def get_model_y_mean_scalers(model, _input):
    """
    Re-casts the sklearn output mean scalers for each output basis

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    mean = [model.Y_scalers[i].mean_[0] for i in range(_input.shape[1])]
    mean = tf.stack(mean, axis=0)
    mean = tf.cast(mean, tf.float32)
    return mean


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
# @tf.function
def y_inv_scale_hack(model, _input):
    """
    Re-creates the sklearn inverse standard scaler as a tensorflow function,

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    y_scaled = _input
    y_scaled = tf.reshape(y_scaled, shape=(-1, y_scaled.shape[1]))

    mean = get_model_y_mean_scalers(model, _input)

    std = get_model_y_std_scalers(model, _input)
    y = y_scaled * std + mean
    return y


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
# @tf.function
def generate_surrogate(x,
                       amp_model,
                       amp_basis,
                       phase_model,
                       phase_basis):
    """
    Predict a full surrogate waveform for a given position and basis

    :param x: The location in parameter space - currently (q, chi1, chi2)
    :param amp_model: The pre-loaded amplitude nn generator
    :param amp_basis: The pre-loaded amplitude basis
    :param phase_model: The pre-loaded phase nn generator
    :param phase_basis: The pre-loaded phase basis
    :return:
    """

    x = tf.transpose(
        tf.stack([tf.math.log(x[:, 0]), x[:, 1], x[:, 2]], axis=0))
    x = tf.cast(x, tf.float64)
    x = tf.reshape(x, shape=(-1, x.shape[0]))
    x = tf.convert_to_tensor(x, dtype=tf.float64)

    amp_alpha = predict_hack(amp_model, x)

    amp_alpha = tf.cast(amp_alpha, tf.float64)
    amp_basis = tf.cast(amp_basis, tf.float64)

    amp = tf.tensordot(amp_alpha,
                       amp_basis,
                       axes=1)

    phase_alpha = predict_hack(phase_model,
                               x)

    phase_alpha = tf.cast(phase_alpha,
                          tf.float32)
    phase_basis = tf.cast(phase_basis,
                          tf.float32)

    phase = tf.tensordot(phase_alpha,
                         phase_basis,
                         axes=1)

    phase = tf.cast(phase, tf.complex64)
    amp = tf.cast(amp, tf.complex64)

    h = amp * tf.math.exp(-1.j * phase)
    phase = tf.cast(phase, tf.float32)
    phase = phase[0]
    phase = phase - phase[0]

    return tf.math.real(h), tf.math.imag(h), amp, phase


def convert_tfp_chains_to_arviz_object(chains):
    """
    Changes the output from sampling into an arviz object, this makes it easier to perform diagnostics
    and plots using arviz

    :param chains: The raw output from the tfp sampler

    :return:
    """
    samples_ar = np.array(chains)
    samples_right_shape = np.swapaxes(samples_ar, 0, 1)
    arv_samples = az.convert_to_inference_data(samples_right_shape)
    return arv_samples



import lal
import lalsimulation as lalsim
import numpy as np
import warnings
import copy

class WaveformRotations(object):
    def __init__(self, times, hlms, frame, *args, **kwargs):
        """
        A class to perform rotations to modes
        times: 1d array of times
        hlms: dict of complex modes
            {(l,m):complex_modes}
        alpha0, thetaJN, phi0: angles to go from L (LAL) inertial frame to J inertial frame
        frame: str choices: ['inertial-L', 'inertial-J', 'coprec']

        if frame is 'inertial-L'
        required_args = ['alpha0', 'thetaJN', 'phi0']
        passed as kwargs

        if frame is 'coprec'
        required_args = ['alpha', 'beta', 'gamma']
        passed as kwargs

        modifies hlms in-place. To find out what frame you are in
        use self.frame.
        """
        frame_choices = ['inertial-L', 'inertial-J', 'coprec']
        if frame not in frame_choices:
            raise ValueError(f"input frame ({frame}) not in frame_choices ({frame_choices})")
        self.frame = frame

        if self.frame == 'inertial-L':
            required_args = ['alpha0', 'thetaJN', 'phi0']
            for r_arg in required_args:
                if r_arg not in kwargs.keys():
                    raise ValueError(f"{r_arg} not in {kwargs.keys()}")
            self.set_alpha0_thetaJN_phi0(kwargs['alpha0'], kwargs['thetaJN'], kwargs['phi0'])

        elif self.frame == 'coprec':
            required_args = ['alpha', 'beta', 'gamma']
            for r_arg in required_args:
                if r_arg not in kwargs.keys():
                    raise ValueError(f"{r_arg} not in {kwargs.keys()}")
            self.set_alpha_beta_gamma(kwargs['alpha'], kwargs['beta'], kwargs['gamma'])

        self.times = times
        self.hlms_original = copy.deepcopy(hlms)
        self.hlms = hlms

    def _rotate(self, times, hlms, r1, r2, r3):
        hlms_3col = convert_to_nrutils_dict(times, hlms)
        hlms_out_3col = {}
        hlms_out = {}
        for k in hlms_3col.keys():
            l, m = k
            hlms_out_3col[k] = rotate_wfarrs_at_all_times(l, m, hlms_3col, [r1, r2, r3])
            hlms_out[k] = hlms_out_3col[k][:,1] + 1.j * hlms_out_3col[k][:,2]

        return hlms_out

    def set_alpha0_thetaJN_phi0(self, alpha0, thetaJN, phi0):
        """
        set rotation angles to go from inertial-L to inertial-J frame
        """
        self.alpha0 = alpha0
        self.thetaJN = thetaJN
        self.phi0 = phi0

    def set_alpha_beta_gamma(self, alpha, beta, gamma):
        """
        set co-precessing angles
        """
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

    def from_inertial_frame_to_j_frame(self):
        assert self.frame == "inertial-L"
        self.hlms = self._rotate(self.times, self.hlms, -self.alpha0 , -self.thetaJN, -self.phi0)
        self.frame = "inertial-J"

    def from_j_frame_to_coprecessing_frame(self, func='new', RoughDirection=None, RoughDirectionIndex=0):
        """
        func: str either 'old' or 'new'.
            'old' for nrutils, 'new' for scri
        RoughDirection: typically either [0,0,1] or [0,0,-1].
            If initial beta angle (thetaJN) is > pi/2 then should be [0,0,-1]
            otherwise use [0,0,1]
        """
        assert self.frame == "inertial-J"
        if func=='old':
            self.alpha, self.beta, self.gamma, self.X, self.Y, self.Z = \
                calc_coprecessing_angles(
                    self.hlms, domain_vals=self.times, return_xyz='all')
        elif func=='new':
            self.alpha, self.beta, self.gamma, self.X, self.Y, self.Z = \
                calc_coprecessing_angles_scri(
                    self.hlms, domain_vals=self.times, RoughDirection=RoughDirection, RoughDirectionIndex=RoughDirectionIndex)
        self.hlms = self._rotate(self.times, self.hlms, -self.gamma , -self.beta, -self.alpha)
        self.frame = "coprec"

    def from_inertial_frame_to_coprecessing_frame(self, func='new', RoughDirection=None, RoughDirectionIndex=0):
        """
        assumes we are in the LAL inertial frame i.e. L along z
        and transforms to the coprecessing frame.
        It first transforms to the J-frame and then
        performs the maximum emission direction calculation

        func: str either 'old' or 'new'.
            'old' for nrutils, 'new' for scri
        """
        assert self.frame == "inertial-L"
        self.from_inertial_frame_to_j_frame()

        if self.thetaJN > np.pi/2:
            RoughDirection = [0,0,-1]
        else:
            RoughDirection = [0,0,1]

        assert self.frame == "inertial-J"
        self.from_j_frame_to_coprecessing_frame(func=func, RoughDirection=RoughDirection, RoughDirectionIndex=RoughDirectionIndex)


    def from_coprecessing_frame_to_j_frame(self):
        """
        assumes we are in the coprecessing frame and performs inverse
        rotations to arrive back at the inerital LAL frame i.e. L along z.
        """
        assert self.frame == "coprec"
        self.hlms = self._rotate(self.times, self.hlms, self.alpha , self.beta, self.gamma)
        self.frame = "inertial-J"

    def from_j_frame_to_inertial_frame(self):
        assert self.frame == "inertial-J"
        if hasattr(self, 'alpha0') is False:
            raise ValueError("please set alpha0")
        if hasattr(self, 'thetaJN') is False:
            raise ValueError("please set thetaJN")
        if hasattr(self, 'phi0') is False:
            raise ValueError("please set phi0")
        self.hlms = self._rotate(self.times, self.hlms, self.phi0 , self.thetaJN, self.alpha0)
        self.frame = "inertial-L"

    def from_coprecessing_frame_to_inertial_frame(self):
        """
        assumes we are in the coprecessing frame and performs inverse
        rotations to arrive back at the inerital LAL frame i.e. L along z.
        """
        assert self.frame == "coprec"
        self.from_coprecessing_frame_to_j_frame()

        assert self.frame == "inertial-J"
        self.from_j_frame_to_inertial_frame()

def compute_L_to_J_angles(mass1, mass2, f_ref, inc, phiref, spin1x, spin1y, spin1z, spin2x, spin2y, spin2z):
    """
    from nrutils
    """
    _, _, _, thetaJN, alpha, phi_aligned, _ = lalsim.SimIMRPhenomPCalculateModelParametersFromSourceFrame(
        mass1*lal.MSUN_SI, mass2*lal.MSUN_SI,
        f_ref, phiref, inc,
        spin1x, spin1y, spin1z,
        spin2x, spin2y, spin2z,
        lalsim.IMRPhenomPv2_V)

    return thetaJN, alpha, phi_aligned


# Calculate Widger D-Matrix Element
def wdelement( ll,         # polar index (eigenvalue) of multipole to be rotated (set of m's for single ll )
               mp,         # member of {all em for |em|<=l} -- potential projection spaceof m
               mm,         # member of {all em for |em|<=l} -- the starting space of m
               alpha,      # -.
               beta,       #  |- Euler angles for rotation
               gamma ):    # -'
    """
    from nrutils
    """

    #** James Healy 6/18/2012
    #** wignerDelement
    #*  calculates an element of the wignerD matrix
    # Modified by llondon6 in 2012 and 2014
    # Converted to python by spxll 2016
    #
    # This implementation apparently uses the formula given in:
    # https://en.wikipedia.org/wiki/Wigner_D-matrix
    #
    # Specifically, this the formula located here: https://wikimedia.org/api/rest_v1/media/math/render/svg/53fd7befce1972763f7f53f5bcf4dd158c324b55

    #
    from numpy import sqrt,exp,cos,sin,ndarray
    from scipy.special import factorial

    #
    if ( (type(alpha) is ndarray) and (type(beta) is ndarray) and (type(gamma) is ndarray) ):
        alpha,beta,gamma = alpha.astype(float), beta.astype(float), gamma.astype(float)
    else:
        alpha,beta,gamma = float(alpha),float(beta),float(gamma)

    #
    coefficient = sqrt( factorial(ll+mp)*factorial(ll-mp)*factorial(ll+mm)*factorial(ll-mm))*exp( 1j*(mp*alpha+mm*gamma) )

    # NOTE that there may be convention differences where the overall sign of the complex exponential may be negated

    #
    total = 0

    # find smin
    if (mm-mp) >= 0 :
        smin = mm - mp
    else:
        smin = 0

    # find smax
    if (ll+mm) > (ll-mp) :
        smax = ll-mp
    else:
        smax = ll+mm

    #
    if smin <= smax:
        for ss in range(smin,smax+1):
            A = (-1)**(mp-mm+ss)
            A *= cos(beta/2)**(2*ll+mm-mp-2*ss)  *  sin(beta/2)**(mp-mm+2*ss)
            B = factorial(ll+mm-ss) * factorial(ss) * factorial(mp-mm+ss) * factorial(ll-mp-ss)
            total += A/B

    #
    element = coefficient*total
    return element

# Given dictionary of multipoles all with the same l, calculate the roated multipole with (l,mp)
def rotate_wfarrs_at_all_times( l,                          # the l of the new multipole (everything should have the same l)
                                m,                          # the m of the new multipole
                                like_l_multipoles_dict,     # dictionary in the format { (l,m): array([domain_values,+,x]) }
                                euler_alpha_beta_gamma,
                                ref_orientation = None ):             #
    """
    from nrutils
    """

    '''
    Given dictionary of multipoles all with the same l, calculate the roated multipole with (l,mp).
    Key reference -- arxiv:1012:2879
    ~ LL,EZH 2018
    '''

    # Import usefuls
    from numpy import exp, pi, array, ones, sign, complex128
#     from nrutils.manipulate.rotate import wdelement

    #
    alpha,beta,gamma = euler_alpha_beta_gamma

    #
    if not ( ref_orientation is None ) :
        raise ValueError('The use of "ref_orientation" has been depreciated for this function.')

    # Handle the default behavior for the reference orientation
    if ref_orientation is None:
        ref_orientation = ones(3)

    # Apply the desired offecrt for the reference orientation. NOTE that this is primarily useful for BAM run which have an atypical coordinate setup if Jz<0
    gamma *= sign( ref_orientation[-1] )
    alpha *= sign( ref_orientation[-1] )

    # Test to see if the original wfarr is complex and if so set the new wfarr to be complex as well
    wfarr_type = type( like_l_multipoles_dict[list(like_l_multipoles_dict.keys())[0]][:,1][0] )

    #
    # new_ylm = 0
    if wfarr_type == complex128:
        new_plus  = 0 + 0j
        new_cross = 0 + 0j
    else:
        new_plus  = 0
        new_cross = 0
    for lm in like_l_multipoles_dict:
        # See eq A9 of arxiv:1012:2879
        l,mp = lm
        old_wfarr = like_l_multipoles_dict[lm]

        #
        # y_mp = old_wfarr[:,1] + 1j*old_wfarr[:,2]
        # new_ylm += wdelement(l,m,mp,alpha,beta,gamma) * y_mp

        #
        d = wdelement(l,m,mp,alpha,beta,gamma)
        a,b = d.real,d.imag
        #
        p = old_wfarr[:,1]
        c = old_wfarr[:,2]
        #
        new_plus  += a*p - b*c
        new_cross += b*p + a*c

    # Construct the new waveform array
    t = old_wfarr[:,0]

    #
    # ans = array( [ t, new_ylm.real, new_ylm.imag ] ).T
    ans = array( [ t, new_plus, new_cross ] ).T

    # Return the answer
    return ans


# Calculate the emission tensor given a dictionary of multipole data
def calc_Lab_tensor( multipole_dict ):
    """
    from nrutils
    """

    '''
    Given a dictionary of multipole moments (single values or time series)
    determine the emission tensor, <L(aLb)>.
    The input must be a dictionary of the format:
    { (2,2):wf_data22, (2,1):wf_data21, ... (l,m):wf_datalm }
    Key referece: https://arxiv.org/pdf/1304.3176.pdf
    Secondary ref: https://arxiv.org/pdf/1205.2287.pdf
    Lionel London 2017
    '''

    # Import usefuls
    from numpy import sqrt,zeros_like,ndarray,zeros,double

    # Rename multipole_dict for short-hand
    y = multipole_dict

    # Allow user to input real and imag parts separately -- this helps with sanity checks
    if isinstance( y[2,2], dict ):
        #
        if not ( ('real' in y[2,2]) and ('imag' in y[2,2]) ):
            raise ValueError('You\'ve entered a multipole dictionary with separate real and imaginary parts. This must be formatted such that y[2,2]["real"] gives the real part and ...')
        #
        x = {}
        lmlist = y.keys()
        for l,m in lmlist:
            x[l,m]        = y[l,m]['real'] + 1j*y[l,m]['imag']
            x[l,m,'conj'] = x[l,m].conj()
    elif isinstance( y[2,2], (float,int,complex,ndarray) ):
        #
        x = {}
        lmlist = y.keys()
        for l,m in lmlist:
            x[l,m]        = y[l,m]
            x[l,m,'conj'] = y[l,m].conj()
    #
    y = x


    # Check type of dictionary values and pre-allocate output
    if isinstance( y[2,2], (float,int,complex) ):
        L = zeros( (3,3), dtype=complex )
    elif isinstance( y[2,2], ndarray ):
        L = zeros( (3,3,len(y[2,2])), dtype=complex )
    else:
        raise ValueError('Dictionary values of handled type; must be float or array')

    # define lambda function for useful coeffs
    c = lambda l,m: sqrt( l*(l+1) - m*(m+1) ) if abs(m)<=l else 0

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#
    # Compute tensor elements (Eqs. A1-A2 of https://arxiv.org/pdf/1304.3176.pdf)
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#

    # Pre-allocate elements
    I0,I1,I2,Izz = zeros_like(y[2,2]), zeros_like(y[2,2]), zeros_like(y[2,2]), zeros_like(y[2,2])

    # Sum contributions from input multipoles
    for l,m in lmlist:

        # Eq. A2c
        I0 += 0.5 * ( l*(l+1)-m*m ) * y[l,m] * y[l,m,'conj']

        # Eq. A2b
        I1 += c(l,m) * (m+0.5) * ( y[l,m+1,'conj'] if (l,m+1) in y else 0 ) * y[l,m]

        # Eq. A2a
        I2 += 0.5 * c(l,m) * c(l,m+1) * y[l,m] * ( y[l,m+2,'conj'] if (l,m+2) in y else 0 )

        # Eq. A2d
        Izz += m*m * y[l,m] * y[l,m,'conj']

    # Compute the net power (amplitude squared) of the multipoles
    N = sum( [ y[l,m] * y[l,m,'conj'] for l,m in lmlist ] ).real

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#
    # Populate the emission tensor ( Eq. A2e )
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#

    # Populate asymmetric elements
    L[0,0] = I0 + I2.real
    L[0,1] = I2.imag
    L[0,2] = I1.real
    L[1,1] = I0 - I2.real
    L[1,2] = I1.imag
    L[2,2] = Izz
    # Populate symmetric elements
    L[1,0] = L[0,1]
    L[2,0] = L[0,2]
    L[2,1] = L[1,2]

    # Normalize
    N[ N==0 ] = min( N[N>0] )
    L = L.real / N

    #
    return L

# Look for point reflections in vector and correct
def reflect_unwrap( vec ):
    """
    from positive.maths
    """
    '''Look for point reflections in vector and correct'''

    #
    from numpy import array,sign,zeros_like

    #
    ans = array(vec)

    #
    for k,v in enumerate(vec):

        #
        if (k>0) and ( (k+1) < len(vec) ):

            #
            l = vec[k-1]
            c = vec[k]
            r = vec[k+1]

            #
            apply_reflection = (sign(l)==sign(r)) and (sign(l)==-sign(c))
            if apply_reflection:

                #
                ans[k] *= -1

    #
    return ans


# Return the min and max limits of an 1D array
def lim(x,dilate=0):
    """
    from positive.maths
    """

    '''
    Return the min and max limits of an 1D array.
    INPUT
    ---
    x,              ndarray
    dilate=0,       fraction of max-min by which to expand or contract output
    RETURN
    ---
    array with [min(x),max(x)]
    '''

    # Import useful bit
    from numpy import array,amin,amax,ndarray,diff

    # ensure is array
    if not isinstance(x,ndarray): x = array(x)

    # Columate input.
    z = x.reshape((x.size,))

    #
    ans = array([min(z),max(z)]) + (0 if len(z)>1 else array([-1e-20,1e-20]))

    #
    if dilate != 0: ans += diff(ans)*dilate*array([-1,1])

    # Return min and max as list
    return ans


# Clone of MATLAB's find function: find all of the elements in a numpy array that satisfy a condition.
def find( bool_vec ):
    """
    from positive.maths
    """

    #
    from numpy import where

    #
    return where(bool_vec)[0]

def spline_diff(t,y,k=3,n=1):
    """
    from positive.maths
    """
    '''
    Wrapper for InterpolatedUnivariateSpline derivative function
    '''

    #
    from numpy import sum
    from scipy.interpolate import InterpolatedUnivariateSpline as spline

    # Calculate the desired number of derivatives
    ans = spline(t,y.real,k=k).derivative(n=n)(t) \
          + ( 1j*spline(t,y.imag,k=k).derivative(n=n)(t) if (sum(abs(y.imag))!=0) else 0 )

    return ans

#
def spline_antidiff(t,y,k=3,n=1):
    """
    from positive.maths
    """
    '''
    Wrapper for InterpolatedUnivariateSpline antiderivative function
    '''

    #
    from scipy.interpolate import InterpolatedUnivariateSpline as spline

    # Calculate the desired number of integrals
    ans = spline(t,y.real,k=k).antiderivative(n=n)(t) + ( 1j*spline(t,y.imag,k=k).antiderivative(n=n)(t) if isinstance(y[0],complex) else 0 )

    # Return the answer
    return ans


# Given a dictionary of multipole data, calculate the Euler angles corresponding to a co-precessing frame
def calc_coprecessing_angles( multipole_dict,       # Dict of multipoles { ... l,m:data_lm ... }
                              domain_vals = None,   # The time or freq series for multipole data
                              ref_orientation = None, # e.g. initial J; used for breaking degeneracies in calculation
                              return_xyz = False,
                              safe_domain_range = None,
                              verbose = None ):
    """
    from nrutils
    """

    '''
    Given a dictionary of multipole data, calculate the Euler angles corresponding to a co-precessing frame
    Key referece: https://arxiv.org/pdf/1304.3176.pdf
    Secondary ref: https://arxiv.org/pdf/1205.2287.pdf
    INPUT
    ---
    multipole_dict,       # dict of multipoles { ... l,m:data_lm ... }
    t,                    # The time series corresponding to multipole data; needed
                            only to calculate gamma; Optional
    verbose,              # Toggle for verbosity
    OUTPUT
    ---
    alpha,beta,gamma euler angles as defined in https://arxiv.org/pdf/1205.2287.pdf
    AUTHOR
    ---
    Lionel London (spxll) 2017
    '''

    # Import usefuls
    from scipy.linalg import eig,norm
    from numpy import arctan2,sin,arcsin,pi,ones,arccos,double,array
    from numpy import unwrap,argmax,cos,array,sqrt,sign,argmin,round

    # Handle optional input
    # if ref_orientation is None: ref_orientation = ones(3)
    if ref_orientation is None: ref_orientation = np.array([0,0,1])

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#
    # Enforce that multipole data is array typed with a well defined length
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#
    y = multipole_dict
    for l,m in y:
        if isinstance( y[l,m], (float,int) ):
            y[l,m] = array( [ y[l,m], ] )
        else:
            if not isinstance(y[l,m],dict):
                # Some input validation
                if domain_vals is None: raise ValueError( 'Since your multipole data is a series, you must also input the related domain_vals (i.e. times or frequencies) array' )
                if len(domain_vals) != len(y[l,m]): raise ValueError('domain_vals array and multipole data not of same length')


    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#
    # Calculate the emission tensor corresponding to the input data
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#
    L = calc_Lab_tensor( multipole_dict )

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#
    # Compute the eigenvectors and values of this tensor
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-#

    # NOTE that members of L have the same length as each y[l,m]; the latter has been
    # forced to always have a length above

    # Initialize idents for angles. NOTE that gamma will be handled below
    alpha,beta = [],[]
    X,Y,Z = [],[],[]

    #
    # reference_z_scale = None
    old_dom_dex = None

    # For all multipole instances
    ref_x,ref_y,ref_z = None,None,None
    flip_z_convention = False
    for k in range( len(L[0,0,:]) ):

        # Select the emission matrix for this instance, k
        _L = L[:,:,k]

        # Compute the eigen vals and vecs for this instance
        vals,vec = eig( _L )

        # Find the dominant direction's index
        dominant_dex = argmax( vals )
        if old_dom_dex is None: old_dom_dex = dominant_dex
        if old_dom_dex != dominant_dex:
            # print dominant_dex
            old_dom_dex = dominant_dex

        # Select the corresponding vector
        dominant_vec = vec[ :, dominant_dex ]

        # There is a z axis degeneracy that we will break here
        # by imposing that the z component is always positive
        # NOTE the following exceptions:
        #  * The optimal emission direction crosses the x-y plane
        #  * Data is processed in a frequency-like domain where f<0 may have z<0

        # if not flip_z_convention:
        #     if sign(dominant_vec[-1]) == -sign(ref_orientation[-1]): dominant_vec *= -1
        # else:
        #     if sign(dominant_vec[-1]) ==  sign(ref_orientation[-1]): dominant_vec *= -1

        # dominant_vec *= sign(domain_vals[k])*sign(ref_orientation[-1])
        # if sign(dominant_vec[-1]) == -sign(ref_orientation[-1]): dominant_vec *= -1

        if not flip_z_convention:
            if sign(dominant_vec[-1]) == -sign(ref_orientation[-1]): dominant_vec *= -1
        else:
            if sign(dominant_vec[-1]) ==  sign(ref_orientation[-1]): dominant_vec *= -1

        # dominant_vec *= sign(domain_vals[k])

        # Extract the components of the dominant eigenvector
        _x,_y,_z = dominant_vec

        # Store reference values if they are None
        if ref_x==None:
            ref_x = _x
            ref_y = _y
            ref_z = _z
        else:
            if (ref_x*_x < 0) and (ref_y*_y < 0):
                _x *= -1
                _y *= -1
                _x *= -1

        # Store unit components for reference in the next iternation
        ref_x = _x
        ref_y = _y
        ref_z = _z

        # Look for and handle trivial cases
        if abs(_x)+abs(_y) < 1e-8 :
            _x = _y = 0

        #
        X.append(_x);Y.append(_y);Z.append(_z)

    # Look for point reflection in X
    X = reflect_unwrap(array(X))
    Y = array(Y)
    Z = array(Z)

    # 3-point vector reflect unwrapping
    # print safe_domain_range
    tol = 0.1
    if safe_domain_range is None: safe_domain_range = lim(abs(domain_vals))
    safe_domain_range = array( safe_domain_range )
    from numpy import arange,mean
    for k in range(len(X))[1:-1]:
        if k>0 and k<(len(domain_vals)-1):

            if (abs(domain_vals[k])>min(abs(safe_domain_range))) and (abs(domain_vals[k])<max(abs(safe_domain_range))):

                left_x_has_reflected = abs(X[k]+X[k-1])<tol*abs(X[k-1])
                left_y_has_reflected = abs(Y[k]+Y[k-1])<tol*abs(X[k-1])

                right_x_has_reflected = abs(X[k]+X[k+1])<tol*abs(X[k])
                right_y_has_reflected = abs(Y[k]+Y[k+1])<tol*abs(X[k])

                x_has_reflected = right_x_has_reflected or left_x_has_reflected
                y_has_reflected = left_y_has_reflected or right_y_has_reflected

                if x_has_reflected and y_has_reflected:

                    # print domain_vals[k]

                    if left_x_has_reflected:
                        X[k:] *=-1
                    if right_x_has_reflected:
                        X[k+1:] *= -1

                    if left_y_has_reflected:
                        Y[k:] *=-1
                    if right_y_has_reflected:
                        Y[k+1:] *= -1

                    Z[k:] *= -1


    # # Enforce that the initial direction of Z is the same as the input reference orientation
    index_ref = find( domain_vals>min(abs(safe_domain_range)) )[0]
    Z_ref = Z[ index_ref ]
    # print 'len(domain_vals) = ',len(domain_vals)
    # print 'index_ref = ',index_ref
    # print 'Z_ref = ',Z_ref
    # print 'ref_orientation = ',ref_orientation
    if Z_ref != 0:
        if sign(Z_ref) != sign( ref_orientation[-1] ):
            warnings.warn("Reference orientation and calculated data inconsistent. We will reflect.")
            Z *= -1
            Y *= -1
            X *= -1

    # Make sure that imag parts are gone
    X = double(X)
    Y = double(Y)
    Z = double(Z)

    #################################################
    # Reflect Y according to nrutils conventions    #
    Y = -Y                                          #
    #################################################

    a = array(ref_orientation)/norm(ref_orientation)
    B = array([X,Y,Z]).T
    b = (B.T/norm(B,axis=1)).T
    xb,yb,zb = b.T
    test_quantity = a[0]*xb+a[1]*yb+a[2]*zb

    mask = (domain_vals>=min(safe_domain_range)) & (domain_vals<=max(safe_domain_range))
    if 1*(test_quantity[mask][0])<0:
        warnings.warn('flipping manually for negative domain')
        X = -X
        Y = -Y
        Z = -Z

    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    #                          Calculate Angles                           #
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

    alpha = arctan2(Y,X)
    beta  = arccos(Z)

    # Make sure that angles are unwrapped
    alpha = unwrap( alpha )
    beta  = unwrap( beta  )

    # Calculate gamma (Eq. A4 of of arxiv:1304.3176)
    if len(alpha) > 1 :
        k = 1
        # NOTE that spline_diff and spline_antidiff live in positive.maths
        gamma = - spline_antidiff( domain_vals, cos(beta) * spline_diff(domain_vals,alpha,k=k), k=k  )
        gamma = unwrap( gamma )
    else:
        # NOTE that this is the same as above, but here we're choosing an integration constant such that the value is zero. Above, no explicit integration constant is chosen.
        gamma = 0

    # Return answer
    if return_xyz == 'all':
        #
        return alpha,beta,gamma,X,Y,Z
    elif return_xyz:
        #
        return X,Y,Z
    else:
        return alpha,beta,gamma

def calc_coprecessing_angles_scri(multipole_dict, domain_vals, RoughDirection=None, RoughDirectionIndex=0):
    from scipy.linalg import eigh
    Lab = calc_Lab_tensor(multipole_dict)
    val, vec = [], []
    for i in range(Lab.shape[2]):
        _val, _vec = eigh(Lab[:,:,i])
        val.append(_val)
        vec.append(_vec)
    val = np.array(val)
    vec = np.array(vec)
    # dpa = eigenvecs[:, :, 2]  # `eigh` always returns eigenvals in *increasing* order
    if RoughDirection is None:
        RoughDirection=np.array([0,0,1])
    else:
        # enforce it to be numpy.ndarray
        RoughDirection = np.array(RoughDirection)
    dpa = vec[:,:,2]
    dpa = _LLDominantEigenvector(dpa, dpa_i=RoughDirection, i_index=RoughDirectionIndex)
    X, Y, Z = dpa.T

    # alpha needed a minus sign for convention reasons...
    alpha = -np.unwrap(np.arctan2(Y,X))
    beta  = np.arccos(Z)
    k = 1
    # NOTE that spline_diff and spline_antidiff live in positive.maths
    gamma = - spline_antidiff( domain_vals, np.cos(beta) * spline_diff(domain_vals,alpha,k=k), k=k  )
    gamma = np.unwrap( gamma )

    return alpha, beta, gamma, X, Y, Z


def _LLDominantEigenvector(dpa, dpa_i, i_index):
    """Taken from https://github.com/moble/scri/blob/master/mode_calculations.py

    Now we find and normalize the dominant principal axis (dpa) at each
    moment, made continuous
    """
    dpa = dpa.copy()
    # Make the initial direction closer to RoughInitialEllDirection than not
    if (
        dpa_i[0] * dpa[i_index, 0]
        + dpa_i[1] * dpa[i_index, 1]
        + dpa_i[2] * dpa[i_index, 2]
    ) < 0.0:
        dpa[i_index, 0] *= -1
        dpa[i_index, 1] *= -1
        dpa[i_index, 2] *= -1
    # Now, go through and make the vectors reasonably continuous.
    d = -1
    LastNorm = np.sqrt(dpa[i_index, 0] ** 2 + dpa[i_index, 1] ** 2 + dpa[i_index, 2] ** 2)
    for i in range(i_index - 1, -1, -1):
        Norm = dpa[i, 0] ** 2 + dpa[i, 1] ** 2 + dpa[i, 2] ** 2
        dNorm = (
            (dpa[i, 0] - dpa[i - d, 0]) ** 2
            + (dpa[i, 1] - dpa[i - d, 1]) ** 2
            + (dpa[i, 2] - dpa[i - d, 2]) ** 2
        )
        if dNorm > Norm:
            dpa[i, 0] *= -1
            dpa[i, 1] *= -1
            dpa[i, 2] *= -1
        # While we're here, let's just normalize that last one
        if LastNorm != 0.0 and LastNorm != 1.0:
            dpa[i - d, 0] /= LastNorm
            dpa[i - d, 1] /= LastNorm
            dpa[i - d, 2] /= LastNorm
        LastNorm = np.sqrt(Norm)
    if LastNorm != 0.0 and LastNorm != 1.0:
        dpa[0, 0] /= LastNorm
        dpa[0, 1] /= LastNorm
        dpa[0, 2] /= LastNorm
    d = 1
    LastNorm = np.sqrt(dpa[i_index, 0] ** 2 + dpa[i_index, 1] ** 2 + dpa[i_index, 2] ** 2)
    for i in range(i_index + 1, dpa.shape[0]):
        Norm = dpa[i, 0] ** 2 + dpa[i, 1] ** 2 + dpa[i, 2] ** 2
        dNorm = (
            (dpa[i, 0] - dpa[i - d, 0]) ** 2
            + (dpa[i, 1] - dpa[i - d, 1]) ** 2
            + (dpa[i, 2] - dpa[i - d, 2]) ** 2
        )
        if dNorm > Norm:
            dpa[i, 0] *= -1
            dpa[i, 1] *= -1
            dpa[i, 2] *= -1
        # While we're here, let's just normalize that last one
        if LastNorm != 0.0 and LastNorm != 1.0:
            dpa[i - d, 0] /= LastNorm
            dpa[i - d, 1] /= LastNorm
            dpa[i - d, 2] /= LastNorm
        LastNorm = np.sqrt(Norm)
    if LastNorm != 0.0 and LastNorm != 1.0:
        dpa[-1, 0] /= LastNorm
        dpa[-1, 1] /= LastNorm
        dpa[-1, 2] /= LastNorm
    return dpa

def convert_to_nrutils_dict(times, hlms):
    """
    convert input times and hlms into a format that nrutils
    based function 'rotate_wfarrs_at_all_times' can read.
    times: numpy array of times
    hlms: dict with keys [(2,2), (2,1), (2,0), ...]
        containing 1d array of complex numbers
    returns:
    hlms_3col: dict with same keys as 'hlms' but now values
        are 3 col arrays of [time, re(hlm), im(hlm)]
    """
    hlms_3col = {}
    for k in hlms.keys():
        hlms_3col[k] = np.array([times, np.real(hlms[k]), np.imag(hlms[k])]).T
    return hlms_3col



#!/usr/bin/env python

from distutils.core import setup
from setuptools import find_packages

# https://stackoverflow.com/questions/26900328/install-dependencies-from-setup-py
import os
thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = thelibFolder + '/requirements.txt'
#install_requires = []  # Examples: ["gunicorn", "docutils>=0.3", "lxml==0.5a7"]
if os.path.isfile(requirementPath):
   with open(requirementPath) as f:
       install_requires = f.read().splitlines()

setup(name='scrinet',
      version='0.0.1',
      description='Modelling',
      author='Sebastian Khan',
      author_email='sebastian.khan@LIGO.org',
      packages=find_packages(),
      install_requires=install_requires,
      url='https://gitlab.com/SpaceTimeKhantinuum/scrinet',
      scripts=[
          'bin/pipeline/scrinet_build_rb',
          'bin/pipeline/scrinet_evaluate_model',
          'bin/pipeline/scrinet_evaluate_coprec_model',
          'bin/pipeline/scrinet_evaluate_7d_coprec_model',
          'bin/pipeline/scrinet_gen_ts_data',
          'bin/pipeline/scrinet_gen_wf_data_3d_non_prec',
          'bin/pipeline/scrinet_combine_wf_data',
          'bin/pipeline/scrinet_fit',
          'bin/pipeline/scrinet_gen_wf_data_2d_non_prec',
          'bin/pipeline/scrinet_gen_wf_data_non_spinning',
          'bin/pipeline/scrinet_gen_wf_data_3d_prec_single_spin',
          'bin/pipeline/scrinet_gen_wf_data_3d_prec_single_spin_coprec',
          'bin/pipeline/scrinet_gen_wf_data_7d_prec_single_spin_coprec',
          'bin/pipeline/scrinet_make_workflow',
          'bin/pipeline/scrinet_make_webpage'
      ]
      )
